/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 stvextrap-openmp.c tiffread.c tiffwrite.c shannon_tools.c -lfftw3_omp -lfftw3 -lm -ltiff -fopenmp -o stvextrap-openmp

  References :

  [1] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [2] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <omp.h>
#include "fftw3.h"

#define VERSION "1.0"

#define FREE_ALL() {					\
    if(NULL != u) free(u);				\
    if(NULL != gx) free(gx);				\
    if(NULL != gy) free(gy);				\
    if(NULL != px) free(px);				\
    if(NULL != py) free(py);				\
    if(NULL != u0) free(u0);				\
    if(NULL != ubar) free(ubar);			\
    if(NULL != mask) free(mask);			\
    if(NULL != mask_tmp) free(mask_tmp);		\
    if(NULL != dft_u) fftw_free(dft_u);			\
    if(NULL != dft_u0) fftw_free(dft_u0);		\
    if(NULL != dft_gx) fftw_free(dft_gx);		\
    if(NULL != dft_gy) fftw_free(dft_gy);		\
    if(NULL != dft_px) fftw_free(dft_px);		\
    if(NULL != dft_py) fftw_free(dft_py);		\
    if(NULL != dft_div) fftw_free(dft_div);		\
    if(NULL != dft_ubar) fftw_free(dft_ubar);		\
    if(NULL != plan_ubar) fftw_destroy_plan(plan_ubar); \
    if(NULL != plan_u0) fftw_destroy_plan(plan_u0);	\
    if(NULL != plan_u) fftw_destroy_plan(plan_u);	\
    if(NULL != plan_px) fftw_destroy_plan(plan_px);	\
    if(NULL != plan_py) fftw_destroy_plan(plan_py);	\
    if(NULL != iplan_u) fftw_destroy_plan(iplan_u);	\
    if(NULL != iplan_gx) fftw_destroy_plan(iplan_gx);	\
    if(NULL != iplan_gy) fftw_destroy_plan(iplan_gy);	\
    if(NULL != iplan_div) fftw_destroy_plan(iplan_div); \
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

/* eval_energy : compute E(u) = STVn(u) */
static double eval_energy(gx,gy,n,Omega_n)
     double *gx,*gy;
     int n,Omega_n;
{
  int adr;
  double E=0.;

#pragma omp parallel for private(adr) reduction(+:E)
  for(adr=0;adr<Omega_n;adr++) E += hypot(gx[adr],gy[adr]);

  return E/((double)n*(double)n);
}


/* update_primal : perform primal step  */
static void update_primal(u,ubar,mask,dft_u,dft_u0,div,plan_u,iplan_u,n,tau,Omega,nx,ny)
     double *u,*ubar,*div,tau,Omega;
     fftw_complex *dft_u,*dft_u0;
     fftw_plan plan_u,iplan_u;
     char *mask;
     int n,nx,ny;
{
  double cof;
  int adr,half_nx;

  half_nx = nx/2; // integer division
  cof = tau/((double)n*(double)n);

  /* set ubar = -theta*u and u = u + (tau/n^2)*div_n(px,py) */
#pragma omp parallel for private(adr)
  for(adr=0;adr<(int)Omega;adr++) {
    ubar[adr] = -u[adr];
    u[adr] += cof*div[adr];
  }

  /* set dft_u = DFT(u) */
  fftw_execute(plan_u);

  /* set dft_u(a,b) = dft_u0(a,b) if mask(a,b) = 1 */
#pragma omp parallel for private(adr)
  for(adr=0;adr<(1+half_nx)*ny;adr++)
    if(mask[adr]) {
      dft_u[adr][0] = dft_u0[adr][0];
      dft_u[adr][1] = dft_u0[adr][1];
  }

  /* set u = Omega * (inverse DFT of dft_u) */
  fftw_execute(iplan_u); // reminder: this call destroys dft_u

  /* set u = u/Omega and finish ubar update */
#pragma omp parallel for private(adr)
  for(adr=0;adr<(int)Omega;adr++) {
    u[adr] /= Omega;
    ubar[adr] += 2.*u[adr];
  }
}


/* update_dual : perform dual step  */
static void update_dual(px,py,gx,gy,n,sigma,Omega_n)
     double *px,*py,*gx,*gy,sigma,Omega_n;
     int n;
{
  int adr;
  double nrm,cof = sigma/((double)n*(double)n);

#pragma omp parallel for private(adr,nrm)
  for(adr=0;adr<(int)Omega_n;adr++) {
    px[adr] += cof*gx[adr];
    py[adr] += cof*gy[adr];
    nrm = px[adr]*px[adr] + py[adr]*py[adr];
    if(nrm > 1.) {
      nrm = sqrt(nrm);
      px[adr] /= nrm;
      py[adr] /= nrm;
    }
  }
}


static void display_usage()
{
  printf("\nSTV-based spectrum extrapolation.\n\n");
  printf("Usage: stvextrap-openmp [-T numTh] [-N Niter] [-n oversampling] [-t tau] [-s sigma] [-f] [-v] [-vv] in mask out\n\n");
  printf("   -T numTh        : (default = 2) number of threads (OpenMP)\n");
  printf("   -N Niter        : (default = 100) number of iterations\n");
  printf("   -n oversampling : (default = 2) oversampling factor (use the STV_n regularizer)\n");
  printf("   -t tau          : primal step for the Chambolle-Pock Algorithm\n");
  printf("   -s sigma        : dual step for the Chambolle-Pock Algorithm\n");
  printf("   -f              : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   -v              : light verbose mode (display progression)\n");
  printf("   -vv             : heavy verbose mode (compute & display energy evolution)\n");
  printf("   in              : input image (TIFF)\n");
  printf("   mask            : binary mask in the Fourier domain (TIFF)\n");
  printf("   out             : output image (TIFF)\n\n");
}


int main(int argc, char **argv)
{
  char *mask = NULL,*filenamein = NULL,*filenamemask = NULL,*filenameout = NULL,*Nvalue = NULL,*nvalue = NULL,*tvalue = NULL,*svalue = NULL,*Tvalue = NULL,vflag,vvflag,fflag;
  double *mask_tmp = NULL, *u0 = NULL,*u = NULL,*ubar = NULL,*px = NULL,*py = NULL,*gx = NULL,*gy = NULL,*div = NULL,sigma,tau,L,Omega,Omega_n,E;
  fftw_complex *dft_div = NULL,*dft_ubar = NULL,*dft_px = NULL,*dft_py = NULL,*dft_gx = NULL,*dft_gy = NULL,*dft_u0 = NULL,*dft_u = NULL;
  fftw_plan plan_u = NULL,plan_u0 = NULL,plan_ubar = NULL,plan_px = NULL,plan_py = NULL,iplan_gx = NULL,iplan_gy = NULL,iplan_div = NULL,iplan_u = NULL;
  int a,b,n,niter,i,err,nx,ny,Nx,Ny,half_nx,half_ny,half_Nx,half_Ny,numTh,nx_tmp,ny_tmp;
  vvflag = vflag = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-v") == 0) vflag = 1;
    else if (strcmp(argv[i],"-vv") == 0) vvflag = 1;
    else if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (strcmp(argv[i],"-T") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -T requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	Tvalue = argv[i+1]; err = sscanf(Tvalue,"%d",&numTh); i++;
	if(err != 1) { display_usage();  printf("Error: could not retrieve properly the argument of option -T\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-N") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -N requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	Nvalue = argv[i+1]; err = sscanf(Nvalue,"%d",&niter); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -N\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-n") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -n requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	nvalue = argv[i+1]; err = sscanf(nvalue,"%d",&n); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -n\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-t") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -t requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	tvalue = argv[i+1]; err = sscanf(tvalue,"%lf",&tau); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -t\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-s") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -s requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	svalue = argv[i+1]; err = sscanf(svalue,"%lf",&sigma); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -s\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (NULL == filenamein) filenamein = argv[i];
    else if (NULL == filenamemask) filenamemask = argv[i];
    else if (NULL == filenameout) filenameout = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* set default values */
  if(NULL == Nvalue) niter = 100;
  if(NULL == nvalue) n = 2;
  if((NULL == svalue) || (NULL == tvalue)) {
    L = M_PI*sqrt(2)*(double)n;
    if(NULL == tvalue) { tau = 0.99/L; }
    if(NULL == svalue) { sigma = 0.99/L; }
  }
  if(NULL == Tvalue) numTh = 2;

  /* Open MP */
  omp_set_num_threads(numTh);
  fftw_init_threads();
  fftw_plan_with_nthreads(numTh);

  /* check consistency */
  if ((NULL == filenamein) || (NULL == filenamemask) || (NULL == filenameout)) {
    display_usage();
    if (NULL == filenamein) printf("Error: input 'in' missing\n\n");
    else if(NULL == filenamemask) printf("Error: input 'mask' missing\n\n");
    else if(NULL == filenameout) printf("Error: output 'out' missing\n\n");
    return EXIT_FAILURE;
  }

  /* deal with verbose mode */
  if(vflag || vvflag) {
    printf("\nInput parameters\n");
    printf("================\n");
    printf("number of iterations: Niter = %d\n",niter);
    printf("oversampling factor: n = %d\n",n);
    printf("primal step for Chambolle-Pock Algorithm: tau = %.17g\n",tau);
    printf("dual step for Chambolle-Pock Algorithm: sigma = %.17g\n\n",sigma);
  }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(vflag || vvflag) {
    printf("Load input images and allocate memory\n");
    printf("=====================================\n");
  }

  if(EXIT_FAILURE == read_tiff_into_double(&u0,&nx,&ny,filenamein,vflag||vvflag)) {
    if(vflag || vvflag) printf("Free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  if(EXIT_FAILURE == read_tiff_into_double(&mask_tmp,&nx_tmp,&ny_tmp,filenamemask,vflag||vvflag)) {
    if(vflag || vvflag) printf("Free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  if ((nx != nx_tmp) || (ny != ny_tmp)) {
    printf("\nError: input images 'in' and 'mask' must have the same dimensions.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  Nx = n*nx; Ny = n*ny; Omega = nx*ny; Omega_n = Nx*Ny;
  half_nx = nx/2; half_ny = ny/2; half_Nx = Nx/2; half_Ny = Ny/2;

  if(vflag || vvflag) { printf("memory allocation: in progress"); fflush(stdout);}

  ASSERT_ALLOC(u    = (double*) malloc(Omega * sizeof(double)));
  ASSERT_ALLOC(ubar = (double*) malloc(Omega * sizeof(double)));
  ASSERT_ALLOC(div  = (double*) malloc(Omega * sizeof(double)));
  ASSERT_ALLOC(gx   = (double*) malloc(Omega_n * sizeof(double)));
  ASSERT_ALLOC(gy   = (double*) malloc(Omega_n * sizeof(double)));
  ASSERT_ALLOC(px   = (double*) calloc(Omega_n , sizeof(double)));
  ASSERT_ALLOC(py   = (double*) calloc(Omega_n , sizeof(double)));
  ASSERT_ALLOC(mask = (char*)   malloc((1+half_nx)*ny * sizeof(char)));

  ASSERT_ALLOC(dft_gx   = (fftw_complex*) fftw_malloc((1+half_Nx)*Ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_gy   = (fftw_complex*) fftw_malloc((1+half_Nx)*Ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_px   = (fftw_complex*) fftw_malloc((1+half_Nx)*Ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_py   = (fftw_complex*) fftw_malloc((1+half_Nx)*Ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_u    = (fftw_complex*) fftw_malloc((1+half_nx)*ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_u0   = (fftw_complex*) fftw_malloc((1+half_nx)*ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_ubar = (fftw_complex*) fftw_malloc((1+half_nx)*ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_div  = (fftw_complex*) fftw_malloc((1+half_nx)*ny * sizeof(fftw_complex)));

  ASSERT_ALLOC(plan_ubar = fftw_plan_dft_r2c_2d(ny,nx,ubar,dft_ubar,FFTW_MEASURE));
  ASSERT_ALLOC(plan_u0   = fftw_plan_dft_r2c_2d(ny,nx,u0,dft_u0,FFTW_ESTIMATE)); // WARNING: choosing FFTW_MEASURE would destroy u0
  ASSERT_ALLOC(plan_u    = fftw_plan_dft_r2c_2d(ny,nx,u,dft_u,FFTW_MEASURE));
  ASSERT_ALLOC(plan_px   = fftw_plan_dft_r2c_2d(Ny,Nx,px,dft_px,FFTW_MEASURE));
  ASSERT_ALLOC(plan_py   = fftw_plan_dft_r2c_2d(Ny,Nx,py,dft_py,FFTW_MEASURE));
  ASSERT_ALLOC(iplan_div = fftw_plan_dft_c2r_2d(ny,nx,dft_div,div,FFTW_MEASURE));
  ASSERT_ALLOC(iplan_u   = fftw_plan_dft_c2r_2d(ny,nx,dft_u,u,FFTW_MEASURE));
  ASSERT_ALLOC(iplan_gx  = fftw_plan_dft_c2r_2d(Ny,Nx,dft_gx,gx,FFTW_MEASURE));
  ASSERT_ALLOC(iplan_gy  = fftw_plan_dft_c2r_2d(Ny,Nx,dft_gy,gy,FFTW_MEASURE));

  if(vflag || vvflag) { printf("\33[2K\rmemory allocation: done\n\n"); fflush(stdout); }




  /***********************************/
  /* KERNEL: process the input image */
  /***********************************/
  if(vflag || vvflag) {
    printf("Start the image restoration process\n");
    printf("===================================\n");
  }

  /* set u = ubar = u0 */
#pragma omp parallel for private(i)
  for(i=0;i<(int)Omega;i++) u[i] = ubar[i] = u0[i];

  /* fill mask (only half-frequency domain is needed) */
#pragma omp parallel for private(a,b)
  for(b=-half_ny;b<=(ny-1)/2;b++)
    for(a=0;a<=half_nx;a++)
      mask[((b+ny)%ny)*(1+half_nx)+a] = (char) mask_tmp[((b+ny)%ny)*nx+a];

  /* set dft_u0 = DFT(u0) */
  fftw_execute(plan_u0);

  /**** main loop ****/
  for(i=1;i<=niter;i++) {

    /* compute Shannon gradient of ubar into gx and gy */
    compute_sgrad(dft_ubar,dft_gx,dft_gy,plan_ubar,iplan_gx,iplan_gy,nx,ny,Nx,Ny);
    /* update dual variable p = (px,py) */
    update_dual(px,py,gx,gy,n,sigma,Omega_n);
    /* compute Shannon divergence into div */
    compute_sdiv(dft_div,dft_px,dft_py,plan_px,plan_py,iplan_div,nx,ny,Nx,Ny);
    /* update primal variables u and ubar */
    update_primal(u,ubar,mask,dft_u,dft_u0,div,plan_u,iplan_u,n,tau,Omega,nx,ny);

    if(vflag) {
      printf("\33[2K\riteration %d/%d",i,niter);
      fflush(stdout);
    }
    else if(vvflag) {
      compute_sgrad(dft_u,dft_gx,dft_gy,plan_u,iplan_gx,iplan_gy,nx,ny,Nx,Ny);
      E = eval_energy(gx,gy,n,(int)Omega_n);
      printf("iteration %d/%d : E = %.10g\n",i,niter,E);
      fflush(stdout);
    }
  }

  if(vflag || vvflag) printf("\n\n");
  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/




  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if(vflag || vvflag) {
    printf("Save restored image as a TIFF image and free memory\n");
    printf("===================================================\n");
  }

  if (EXIT_FAILURE == save_tiff(filenameout,u,nx,ny,8,fflag,1,vflag)) {
    if (vflag || vvflag) printf("free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /* free memory */
  if(vflag || vvflag) {
    printf("free memory: in progress");
    fflush(stdout);
  }

  FREE_ALL();

  if(vflag || vvflag) {
    printf("\33[2K\rfree memory: done\n\n");
    fflush(stdout);
  }
}
